@@ -1,89 +1,47 @@
 /*
 
-  $Id: Acquisition.c,v 1.8 1997/07/02 20:15:27 jim Exp $
+  Copyright (c) 1996-1998, Digital Creations, Fredericksburg, VA, USA.  
+  All rights reserved.
+  
+  Redistribution and use in source and binary forms, with or without
+  modification, are permitted provided that the following conditions are
+  met:
+  
+    o Redistributions of source code must retain the above copyright
+      notice, this list of conditions, and the disclaimer that follows.
+  
+    o Redistributions in binary form must reproduce the above copyright
+      notice, this list of conditions, and the following disclaimer in
+      the documentation and/or other materials provided with the
+      distribution.
+  
+    o Neither the name of Digital Creations nor the names of its
+      contributors may be used to endorse or promote products derived
+      from this software without specific prior written permission.
+  
+  
+  THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS AND CONTRIBUTORS *AS
+  IS* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
+  TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+  PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL
+  CREATIONS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
+  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
+  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
+  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
+  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
+  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
+  USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
+  DAMAGE.
 
-  Acquisition Wrappers -- Implementation of acquisition through wrappers
-
-
-     Copyright 
-
-       Copyright 1996 Digital Creations, L.C., 910 Princess Anne
-       Street, Suite 300, Fredericksburg, Virginia 22401 U.S.A. All
-       rights reserved.  Copyright in this software is owned by DCLC,
-       unless otherwise indicated. Permission to use, copy and
-       distribute this software is hereby granted, provided that the
-       above copyright notice appear in all copies and that both that
-       copyright notice and this permission notice appear. Note that
-       any product, process or technology described in this software
-       may be the subject of other Intellectual Property rights
-       reserved by Digital Creations, L.C. and are not licensed
-       hereunder.
-
-     Trademarks 
-
-       Digital Creations & DCLC, are trademarks of Digital Creations, L.C..
-       All other trademarks are owned by their respective companies. 
-
-     No Warranty 
-
-       The software is provided "as is" without warranty of any kind,
-       either express or implied, including, but not limited to, the
-       implied warranties of merchantability, fitness for a particular
-       purpose, or non-infringement. This software could include
-       technical inaccuracies or typographical errors. Changes are
-       periodically made to the software; these changes will be
-       incorporated in new editions of the software. DCLC may make
-       improvements and/or changes in this software at any time
-       without notice.
-
-     Limitation Of Liability 
-
-       In no event will DCLC be liable for direct, indirect, special,
-       incidental, economic, cover, or consequential damages arising
-       out of the use of or inability to use this software even if
-       advised of the possibility of such damages. Some states do not
-       allow the exclusion or limitation of implied warranties or
-       limitation of liability for incidental or consequential
-       damages, so the above limitation or exclusion may not apply to
-       you.
-
-    If you have questions regarding this software,
-    contact:
-   
-      Jim Fulton, jim@digicool.com
-      Digital Creations L.C.  
-   
-      (540) 371-6909
-
-
-  Full description
-
-  $Log: Acquisition.c,v $
-  Revision 1.8  1997/07/02 20:15:27  jim
-  Added stupid parens to make 'gcc -Wall -pedantic' and Barry happy.
-
-  Revision 1.7  1997/06/19 19:32:31  jim
-  *** empty log message ***
-
-  Revision 1.6  1997/06/19 19:31:39  jim
-  Added ident string.
-
-  Revision 1.5  1997/06/19 19:24:21  jim
-  Many fixes and consolodation with Xaq.
-
-  Revision 1.4  1997/02/20 00:55:29  jim
-  *** empty log message ***
-
-  Revision 1.3  1997/02/19 22:30:33  jim
-  Added $#@! missing static declaration.
-
-  Revision 1.2  1997/02/17 16:20:11  jim
-  Fixed bug in mix-in class declaration.
-  Added __version__.
-
-  Revision 1.1  1997/02/17 15:05:40  jim
-  *** empty log message ***
+  $Id: Acquisition.c,v 1.24 1998/11/17 19:49:43 jim Exp $
 
+  If you have questions regarding this software,
+  contact:
+ 
+    Digital Creations L.C.  
+    info@digicool.com
+ 
+    (540) 371-6909
 
 */
 #include "ExtensionClass.h"
@@ -110,6 +68,8 @@ static PyObject *py__add__, *py__sub__, *py__mul__, *py__div__,
   *py__concat__, *py__repeat__, *py__len__, *py__of__, *py__call__,
   *py__repr__, *py__str__;
 
+static PyObject *Acquired=0;
+
 static void
 init_py_names()
 {
@@ -175,10 +135,13 @@ typedef struct {
   PyObject *container;
 } Wrapper;
 
+static Wrapper *freeWrappers=0;
+
 staticforward PyExtensionClass Wrappertype, XaqWrappertype;
 
 #define isWrapper(O) ((O)->ob_type==(PyTypeObject*)&Wrappertype || \
 		      (O)->ob_type==(PyTypeObject*)&XaqWrappertype)
+#define WRAPPER(O) ((Wrapper*)(O))
 
 static PyObject *
 Wrapper__init__(Wrapper *self, PyObject *args)
@@ -198,17 +161,43 @@ Wrapper__init__(Wrapper *self, PyObject *args)
 
 /* ---------- */
 
-static Wrapper *
+static PyObject *
+__of__(PyObject *inst, PyObject *parent)
+{
+  PyObject *r, *t;
+
+  UNLESS(r=PyObject_GetAttr(inst, py__of__)) return NULL;
+  UNLESS(t=PyTuple_New(1)) goto err;
+  PyTuple_SET_ITEM(t,0,parent);
+  ASSIGN(r,PyObject_CallObject(r,t));
+  PyTuple_SET_ITEM(t,0,NULL);
+  Py_DECREF(t);
+  return r;
+err:
+  Py_DECREF(r);
+  return NULL;
+}
+
+static PyObject *
 newWrapper(PyObject *obj, PyObject *container, PyTypeObject *Wrappertype)
 {
   Wrapper *self;
   
-  UNLESS(self = PyObject_NEW(Wrapper, Wrappertype)) return NULL;
+  if(freeWrappers)
+    {
+      self=freeWrappers;
+      freeWrappers=(Wrapper*)self->obj;
+      self->ob_type=Wrappertype;
+      self->ob_refcnt=1;
+    }
+  else
+    UNLESS(self = PyObject_NEW(Wrapper, Wrappertype)) return NULL;
+
   Py_INCREF(obj);
   Py_INCREF(container);
   self->obj=obj;
   self->container=container;
-  return self;
+  return OBJECT(self);
 }
 
 
@@ -217,72 +206,277 @@ Wrapper_dealloc(Wrapper *self)
 {
   Py_DECREF(self->obj);
   Py_DECREF(self->container);
-  PyMem_DEL(self);
+  self->obj=OBJECT(freeWrappers);
+  freeWrappers=self;
+}
+
+static PyObject *
+Wrapper_special(Wrapper *self, char *name, PyObject *oname)
+{
+  PyObject *r=0;
+
+  switch(*name)
+    {
+    case 'b':
+      if(strcmp(name,"base")==0)
+	{
+	  if(self->obj)
+	    {
+	      r=self->obj;
+	      while(isWrapper(r) && WRAPPER(r)->obj) r=WRAPPER(r)->obj;
+	    }
+	  else r=Py_None;
+	  Py_INCREF(r);
+	  return r;
+	}
+      break;
+    case 'p':
+      if(strcmp(name,"parent")==0)
+	{
+	  if(self->container) r=self->container;
+	  else r=Py_None;
+	  Py_INCREF(r);
+	  return r;
+	}
+      break;
+    case 's':
+      if(strcmp(name,"self")==0)
+	{
+	  if(self->obj) r=self->obj;
+	  else r=Py_None;
+	  Py_INCREF(r);
+	  return r;
+	}
+      break;
+    case 'a':
+      if(strcmp(name,"acquire")==0)
+	{
+	  return Py_FindAttr(OBJECT(self),oname);
+	}
+      break;
+    case 'i':
+      if(strcmp(name,"inContextOf")==0)
+	{
+	  return Py_FindAttr(OBJECT(self),oname);
+	}
+      break;
+      
+    }
+
+  return NULL;
+}
+
+static int
+apply_filter(PyObject *filter, PyObject *inst, PyObject *oname, PyObject *r,
+	     PyObject *extra, PyObject *orig)
+{
+  PyObject *fr;
+  int ir;
+
+  UNLESS(fr=PyTuple_New(5)) goto err;
+  PyTuple_SET_ITEM(fr,0,orig);
+  Py_INCREF(orig);
+  PyTuple_SET_ITEM(fr,1,inst);
+  Py_INCREF(inst);
+  PyTuple_SET_ITEM(fr,2,oname);
+  Py_INCREF(oname);
+  PyTuple_SET_ITEM(fr,3,r);
+  Py_INCREF(r);
+  PyTuple_SET_ITEM(fr,4,extra);
+  Py_INCREF(extra);
+  UNLESS_ASSIGN(fr,PyObject_CallObject(filter, fr)) goto err;
+  ir=PyObject_IsTrue(fr);
+  Py_DECREF(fr);
+  if(ir) return 1;
+  Py_DECREF(r);
+  return 0;
+err:
+  Py_DECREF(r);
+  return -1;
+}
+
+static PyObject *
+Wrapper_acquire(Wrapper *self, PyObject *oname,
+		PyObject *filter, PyObject *extra, PyObject *orig,
+		int sob, int sco)
+{
+  PyObject *r, *v, *tb;
+  char *name;
+
+  name=PyString_AsString(oname);
+  if(*name++=='a' && *name++=='q' && *name++=='_'
+     && (r=Wrapper_special(self, name, oname))) return r;
+
+  if(sob && self->obj)
+    {
+      if(isWrapper(self->obj))
+	{
+	  r=Wrapper_acquire((Wrapper*)self->obj,
+			    oname,filter,extra,orig,1,1);
+	  if(r) goto acquired;
+	  PyErr_Fetch(&r,&v,&tb);
+	  if(r != PyExc_AttributeError || PyObject_Compare(v,oname) != 0)
+	    {
+	      PyErr_Restore(r,v,tb);
+	      return NULL;
+	    }
+	  r=NULL;
+	}
+      else if((r=PyObject_GetAttr(self->obj,oname)))
+	{
+	  if(isWrapper(r) && WRAPPER(r)->container==self->obj)
+	    {
+	      if(r->ob_refcnt==1)
+		{
+		  Py_INCREF(self);
+		  ASSIGN(((Wrapper*)r)->container,OBJECT(self));
+		}
+	      else
+		ASSIGN(r,newWrapper(((Wrapper*)r)->obj,
+				    OBJECT(self), r->ob_type));
+	    }
+	  else if(PyECMethod_Check(r) && PyECMethod_Self(r)==self->obj)
+	    ASSIGN(r,PyECMethod_New(r,OBJECT(self)));
+	  else if(has__of__(r)) ASSIGN(r,__of__(r,OBJECT(self)));
+	  if(filter)
+	    switch(apply_filter(filter,OBJECT(self),oname,r,extra,orig))
+	      {
+	      case -1: return NULL;
+	      case 1: return r;
+	      }
+	  else return r;
+	}
+      PyErr_Clear();
+    }
+
+  if(sco && self->container) 
+    {
+      if(isWrapper(self->container))
+	{
+	  if(self->obj && self->obj->ob_type == (PyTypeObject*)&Wrappertype)
+	    {
+	      if(WRAPPER(self->obj)->container==
+		 WRAPPER(self->container)->container) sob=1, sco=0;
+	      else if(WRAPPER(self->obj)->container==
+		      WRAPPER(self->container)->obj)  sob=0, sco=1;
+	      else                                    sob=1, sco=1;
+	      r=Wrapper_acquire((Wrapper*)self->container,
+				oname,filter,extra,orig, sob, sco);
+	   }
+	  else
+	    r=Wrapper_acquire((Wrapper*)self->container,
+			      oname,filter,extra,orig,1,1);
+
+	  
+	  if(r) goto acquired;
+	  return NULL;
+	}
+      else
+	{
+	  if((r=PyObject_GetAttr(self->container,oname)))
+	    if(filter)
+	      switch(apply_filter(filter,self->container,oname,r,extra,orig))
+		{
+		case -1: return NULL;
+		case 1: goto acquired;
+		}
+	    else goto acquired;
+	}
+      PyErr_Clear();
+    }
+
+  PyErr_SetObject(PyExc_AttributeError,oname);
+  return NULL;
+
+acquired:
+  if(has__of__(r)) ASSIGN(r,__of__(r,OBJECT(self)));
+  return r;
+}
+
+static PyObject *
+handle_Acquired(Wrapper *self, PyObject *oname, PyObject *r)
+{
+  UNLESS(self->container)
+    {
+      PyErr_SetObject(PyExc_AttributeError, oname);
+      return NULL;
+    }
+  if(isWrapper(self->container))
+    ASSIGN(r,Wrapper_acquire(WRAPPER(self->container),
+			     oname,NULL,NULL,NULL,1,1));
+  else
+    ASSIGN(r,PyObject_GetAttr(self->container,oname));
+  
+  if(r && has__of__(r)) ASSIGN(r, __of__(r,OBJECT(self)));
+  return r;
+}
+
+static PyObject *
+Wrapper_getattro_(Wrapper *self, PyObject *oname, int sob, int sco)
+{
+  PyObject *r;
+  char *name;
+
+  name=PyString_AsString(oname);
+  if(*name=='a' && name[1]=='q' && name[2]=='_'
+     && (r=Wrapper_special(self, name+3, oname))) return r;
+
+  if(sob && self->obj && (r=PyObject_GetAttr(self->obj,oname)))
+    {
+      if(r == Acquired) return handle_Acquired(self, oname, r);
+      if(isWrapper(r) && WRAPPER(r)->container==self->obj)
+	{
+	  if(r->ob_refcnt==1)
+	    {
+	      Py_INCREF(self);
+	      ASSIGN(((Wrapper*)r)->container,OBJECT(self));
+	    }
+	  else
+	    ASSIGN(r, newWrapper(((Wrapper*)r)->obj,OBJECT(self),r->ob_type));
+	}
+      else if(PyECMethod_Check(r) && PyECMethod_Self(r)==self->obj)
+	ASSIGN(r,PyECMethod_New(r,OBJECT(self)));
+      else if(has__of__(r))
+	ASSIGN(r,__of__(r,OBJECT(self)));
+      return r;
+    }
+  if(self->obj) PyErr_Clear();
+
+  if((*name != '_')
+     && self->container && sco)
+    {
+      if(self->container->ob_type == self->ob_type &&
+	 self->obj->ob_type==self->ob_type)
+	{
+	  if(WRAPPER(self->obj)->container==
+	     WRAPPER(self->container)->container) sob=1, sco=0;
+	  else if(WRAPPER(self->obj)->container==
+		  WRAPPER(self->container)->obj)  sob=0, sco=1;
+	  else                                    sob=1, sco=1;
+	  r=Wrapper_getattro_(WRAPPER(self->container), oname, sob, sco);
+	}
+      else r=PyObject_GetAttr(self->container,oname);
+
+      if(r)
+	{
+	  if(has__of__(r))
+	    ASSIGN(r, __of__(r,OBJECT(self)));
+	  return r;
+	}
+      PyErr_Clear();
+    }
+
+  if(*name++=='_' && strcmp(name,"_init__")==0)
+    return Py_FindAttr(OBJECT(self),oname);
+
+  PyErr_SetObject(PyExc_AttributeError,oname);
+  return NULL;
 }
 
 static PyObject *
 Wrapper_getattro(Wrapper *self, PyObject *oname)
 {
-  PyObject *r;
-  char *name;
-
-  if(self->obj && (r=PyObject_GetAttr(self->obj,oname)))
-    {
-      if(r->ob_type==self->ob_type)
-	{
-	  if(r->ob_refcnt==1)
-	    {
-	      Py_INCREF(self);
-	      ASSIGN(((Wrapper*)r)->container,(PyObject*)self);
-	    }
-	  else
-	    ASSIGN(r,(PyObject*)newWrapper(((Wrapper*)r)->obj,
-					   (PyObject*)self,self->ob_type));
-	}
-      else if(PyECMethod_Check(r) && PyECMethod_Self(r)==self->obj)
-	ASSIGN(r,PyECMethod_New(r,(PyObject*)self));
-      else if(has__of__(r))
-	ASSIGN(r,CallMethodO(r,py__of__,Build("(O)", self), NULL));
-      return r;
-    }
-  if(self->obj) PyErr_Clear();
-
-  name=PyString_AsString(oname);
-
-  if(*name != '_')
-    {
-      
-      if(*name++=='a' && *name++=='q' && *name++=='_')
-	{
-	  if(strcmp(name,"parent")==0)
-	    {
-	      if(self->container) r=self->container;
-	      else r=Py_None;
-	      Py_INCREF(r);
-	      return r;
-	    }
-	  if(strcmp(name,"self")==0)
-	    {
-	      if(self->obj) r=self->obj;
-	      else r=Py_None;
-	      Py_INCREF(r);
-	      return r;
-	    }
-	}
-
-      if(self->container) 
-	{
-	  if((r=PyObject_GetAttr(self->container,oname)))
-	    return r;
-	  PyErr_Clear();
-	}
-    }
-
-  if(*name++=='_' && strcmp(name,"_init__")==0)
-    return Py_FindAttr((PyObject*)self,oname);
-
-  PyErr_SetObject(PyExc_AttributeError,oname);
-  return NULL;
+  return Wrapper_getattro_(self, oname, 1, 1);
 }
 
 static PyObject *
@@ -291,126 +485,45 @@ Xaq_getattro(Wrapper *self, PyObject *oname)
   PyObject *r;
   char *name;
 
+  name=PyString_AsString(oname);
+
+  if(*name=='_')
+    {
+      if(strcmp(name,"__init__")==0) return Py_FindAttr(OBJECT(self),oname);
+    }  
+
+  if(*name=='a')
+    {
+      if(name[1]=='c')
+	{
+	  if(strcmp(name+2,"quire")==0)
+	    return Py_FindAttr(OBJECT(self),oname);
+	}      
+      else if(name[1]=='q' && name[2]=='_'
+	      && (r=Wrapper_special(self, name+3, oname))) return r;
+    }
+
   if(self->obj && (r=PyObject_GetAttr(self->obj,oname)))
     {
-      if(r->ob_type==self->ob_type)
+      if(r==Acquired) return handle_Acquired(self,oname,r);
+      if(isWrapper(r) && WRAPPER(r)->container==self->obj)
 	{
 	  if(r->ob_refcnt==1)
 	    {
 	      Py_INCREF(self);
-	      ASSIGN(((Wrapper*)r)->container,(PyObject*)self);
+	      ASSIGN(((Wrapper*)r)->container,OBJECT(self));
 	    }
 	  else
-	    ASSIGN(r,(PyObject*)newWrapper(((Wrapper*)r)->obj,
-					   (PyObject*)self, self->ob_type));
+	    ASSIGN(r, newWrapper(((Wrapper*)r)->obj,
+				 OBJECT(self), r->ob_type));
 	}
       else if(PyECMethod_Check(r) && PyECMethod_Self(r)==self->obj)
-	ASSIGN(r,PyECMethod_New(r,(PyObject*)self));
-      else if(has__of__(r))
-	ASSIGN(r,CallMethodO(r,py__of__,Build("(O)", self), NULL));
+	ASSIGN(r,PyECMethod_New(r,OBJECT(self)));
+      else if(has__of__(r)) ASSIGN(r,__of__(r,OBJECT(self)));
       return r;
     }
   if(self->obj) PyErr_Clear();
 
-  name=PyString_AsString(oname);
-
-  if(*name=='a' && strcmp(name,"acquire")==0)
-    return Py_FindAttr((PyObject*)self,oname);
-
-  if(*name=='_' && strcmp(name,"__init__")==0)
-    return Py_FindAttr((PyObject*)self,oname);
-
-  if(*name++=='a' && *name++=='q' && *name++=='_')
-    {
-      if(strcmp(name,"parent")==0)
-	{
-	  if(self->container) r=self->container;
-	  else r=Py_None;
-	  Py_INCREF(r);
-	  return r;
-	}
-      if(strcmp(name,"self")==0)
-	{
-	  if(self->obj) r=self->obj;
-	  else r=Py_None;
-	  Py_INCREF(r);
-	  return r;
-	}
-    }
-
-  PyErr_SetObject(PyExc_AttributeError,oname);
-  return NULL;
-}
-
-static PyObject *
-Wrapper_acquire(Wrapper *self, PyObject *oname)
-{
-  PyObject *r;
-  char *name;
-
-  if(self->obj && (r=PyObject_GetAttr(self->obj,oname)))
-    {
-      if(r->ob_type==self->ob_type)
-	{
-	  if(r->ob_refcnt==1)
-	    {
-	      Py_INCREF(self);
-	      ASSIGN(((Wrapper*)r)->container,(PyObject*)self);
-	    }
-	  else
-	    ASSIGN(r,(PyObject*)newWrapper(((Wrapper*)r)->obj,
-					   (PyObject*)self, self->ob_type));
-	}
-      else if(PyECMethod_Check(r) && PyECMethod_Self(r)==self->obj)
-	ASSIGN(r,PyECMethod_New(r,(PyObject*)self));
-      else if(has__of__(r))
-	ASSIGN(r,CallMethodO(r,py__of__,Build("(O)", self), NULL));
-      return r;
-    }
-  if(self->obj) PyErr_Clear();
-
-  name=PyString_AsString(oname);
-
-  if(*name != '_')
-    {
-
-      if(*name++=='a' && *name++=='q' && *name++=='_')
-	{
-	  if(strcmp(name,"parent")==0)
-	    {
-	      if(self->container) r=self->container;
-	      else r=Py_None;
-	      Py_INCREF(r);
-	      return r;
-	    }
-	  if(strcmp(name,"self")==0)
-	    {
-	      if(self->obj) r=self->obj;
-	      else r=Py_None;
-	      Py_INCREF(r);
-	      return r;
-	    }
-	}
-
-      if(self->container) 
-	{
-	  if(isWrapper(self->container))
-	    {
-	      if((r=Wrapper_acquire((Wrapper*)self->container,oname)))
-		return r;
-	    }
-	  else
-	    {
-	      if((r=PyObject_GetAttr(self->container,oname)))
-		return r;
-	    }
-	  PyErr_Clear();
-	}
-    }
-
-  if(*name++=='_' && strcmp(name,"_init__")==0)
-    return Py_FindAttr((PyObject*)self,oname);
-
   PyErr_SetObject(PyExc_AttributeError,oname);
   return NULL;
 }
@@ -433,7 +546,7 @@ Wrapper_repr(Wrapper *self)
 {
   PyObject *r;
 
-  if((r=PyObject_GetAttr((PyObject*)self,py__repr__)))
+  if((r=PyObject_GetAttr(OBJECT(self),py__repr__)))
     {
       ASSIGN(r,PyObject_CallFunction(r,NULL,NULL));
       return r;
@@ -450,7 +563,7 @@ Wrapper_str(Wrapper *self)
 {
   PyObject *r;
 
-  if((r=PyObject_GetAttr((PyObject*)self,py__str__)))
+  if((r=PyObject_GetAttr(OBJECT(self),py__str__)))
     {
       ASSIGN(r,PyObject_CallFunction(r,NULL,NULL));
       return r;
@@ -472,11 +585,9 @@ static PyObject *
 Wrapper_call(Wrapper *self, PyObject *args, PyObject *kw)
 {
   Py_INCREF(args);
-  return CallMethodO((PyObject*)self,py__call__,args,kw);
+  return CallMethodO(OBJECT(self),py__call__,args,kw);
 }
 
-
-
 /* Code to handle accessing Wrapper objects as sequence objects */
 
 static int
@@ -485,7 +596,33 @@ Wrapper_length(Wrapper *self)
   long l;
   PyObject *r;
 
-  UNLESS(r=CallMethodO((PyObject*)self,py__len__,NULL,NULL))  return -1;
+  UNLESS(r=PyObject_GetAttr(OBJECT(self), py__len__))
+    {
+      /* Hm. Maybe we are being checked to see if we are true.
+	 
+	 Check to see if we have a __getitem__.  If we don't, then
+	 answer that we are true, otherwise raise an error.
+	 */
+      PyErr_Clear();
+      if((r=PyObject_GetAttr(OBJECT(self), py__getitem__)))
+	{
+	  /* Hm, we have getitem, must be error */
+	  Py_DECREF(r);
+	  PyErr_SetObject(PyExc_AttributeError, py__len__);
+	  return -1;
+	}
+      PyErr_Clear();
+
+      /* Try nonzero */
+      UNLESS(r=PyObject_GetAttr(OBJECT(self), py__nonzero__))
+	{
+	  /* No nonzero, it's true :-) */
+	  PyErr_Clear();
+	  return 1;
+	}
+    }
+  
+  UNLESS_ASSIGN(r,PyObject_CallObject(r,NULL)) return -1;
   l=PyInt_AsLong(r);
   Py_DECREF(r);
   return l;
@@ -494,25 +631,25 @@ Wrapper_length(Wrapper *self)
 static PyObject *
 Wrapper_concat(Wrapper *self, PyObject *bb)
 {
-  return CallMethodO((PyObject*)self,py__concat__,Build("(O)", bb) ,NULL);
+  return CallMethodO(OBJECT(self),py__concat__,Build("(O)", bb) ,NULL);
 }
 
 static PyObject *
 Wrapper_repeat(Wrapper *self, int  n)
 {
-  return CallMethodO((PyObject*)self,py__repeat__,Build("(i)", n),NULL);
+  return CallMethodO(OBJECT(self),py__repeat__,Build("(i)", n),NULL);
 }
 
 static PyObject *
 Wrapper_item(Wrapper *self, int  i)
 {
-  return CallMethodO((PyObject*)self,py__getitem__, Build("(i)", i),NULL);
+  return CallMethodO(OBJECT(self),py__getitem__, Build("(i)", i),NULL);
 }
 
 static PyObject *
 Wrapper_slice(Wrapper *self, int  ilow, int  ihigh)
 {
-  return CallMethodO((PyObject*)self,py__getslice__,
+  return CallMethodO(OBJECT(self),py__getslice__,
 		     Build("(ii)", ilow, ihigh),NULL);
 }
 
@@ -521,14 +658,14 @@ Wrapper_ass_item(Wrapper *self, int  i, PyObject *v)
 {
   if(v)
     {
-      UNLESS(v=CallMethodO((PyObject*)self,py__setitem__,
+      UNLESS(v=CallMethodO(OBJECT(self),py__setitem__,
 			   Build("(iO)", i, v),NULL))
 	return -1;
     }
   else
     {
-      UNLESS(v=CallMethodO((PyObject*)self,py__delitem__,
-			   Build("(iO)", i),NULL))
+      UNLESS(v=CallMethodO(OBJECT(self),py__delitem__,
+			   Build("(i)", i),NULL))
 	return -1;
     }
   Py_DECREF(v);
@@ -540,13 +677,13 @@ Wrapper_ass_slice(Wrapper *self, int  ilow, int  ihigh, PyObject *v)
 {
   if(v)
     {
-      UNLESS(v=CallMethodO((PyObject*)self,py__setslice__,
+      UNLESS(v=CallMethodO(OBJECT(self),py__setslice__,
 			   Build("(iiO)", ilow, ihigh, v),NULL))
 	return -1;
     }
   else
     {
-      UNLESS(v=CallMethodO((PyObject*)self,py__delslice__,
+      UNLESS(v=CallMethodO(OBJECT(self),py__delslice__,
 			   Build("(ii)", ilow, ihigh),NULL))
 	return -1;
     }
@@ -571,7 +708,7 @@ static PySequenceMethods Wrapper_as_sequence = {
 static PyObject *
 Wrapper_subscript(Wrapper *self, PyObject *key)
 {
-  return CallMethodO((PyObject*)self,py__getitem__,Build("(O)", key),NULL);
+  return CallMethodO(OBJECT(self),py__getitem__,Build("(O)", key),NULL);
 }
 
 static int
@@ -579,13 +716,13 @@ Wrapper_ass_sub(Wrapper *self, PyObject *key, PyObject *v)
 {
   if(v)
     {
-      UNLESS(v=CallMethodO((PyObject*)self,py__setitem__,
+      UNLESS(v=CallMethodO(OBJECT(self),py__setitem__,
 			   Build("(OO)", key, v),NULL))
 	return -1;
     }
   else
     {
-      UNLESS(v=CallMethodO((PyObject*)self,py__delitem__,
+      UNLESS(v=CallMethodO(OBJECT(self),py__delitem__,
 			   Build("(O)", key),NULL))
 	return -1;
     }
@@ -601,11 +738,47 @@ static PyMappingMethods Wrapper_as_mapping = {
 
 /* -------------------------------------------------------- */
 
+static PyObject *
+Wrapper_acquire_method(Wrapper *self, PyObject *args)
+{
+  PyObject *name, *filter=0, *extra=Py_None;
+
+  UNLESS(PyArg_ParseTuple(args,"O|OO",&name,&filter,&extra)) return NULL;
+
+  return Wrapper_acquire(self,name,filter,extra,OBJECT(self),1,1);
+}
+
+static PyObject *
+Wrapper_inContextOf(Wrapper *self, PyObject *args)
+{
+  PyObject *o, *c;
+  int inner=0;
+
+  UNLESS(PyArg_ParseTuple(args,"O|i",&o,&inner)) return NULL;
+
+  if(inner)
+    while(self->obj && isWrapper(self->obj))
+      self=WRAPPER(self->obj);
+
+  if(OBJECT(self)==o) return PyInt_FromLong(1);
+  c=self->container;
+  while(1)
+    {
+      if(c==o) return PyInt_FromLong(1);
+      if(isWrapper(c)) c=WRAPPER(c)->container;
+      else return PyInt_FromLong(0);
+    }
+}
+
 static struct PyMethodDef Wrapper_methods[] = {
   {"__init__", (PyCFunction)Wrapper__init__, 0,
    "Initialize an Acquirer Wrapper"},
-  {"acquire", (PyCFunction)Wrapper_acquire, 0,
+  {"acquire", (PyCFunction)Wrapper_acquire_method, METH_VARARGS,
    "Get an attribute, acquiring it if necessary"},
+  {"aq_acquire", (PyCFunction)Wrapper_acquire_method, METH_VARARGS,
+   "Get an attribute, acquiring it if necessary"},
+  {"aq_inContextOf", (PyCFunction)Wrapper_inContextOf, METH_VARARGS,
+   "Test whether the object is currently in the context of the argument"},
   {NULL,		NULL}		/* sentinel */
 };
 
@@ -682,7 +855,7 @@ acquire_of(PyObject *self, PyObject *args)
       return NULL;
     }
 
-  return (PyObject*)newWrapper(self,args,(PyTypeObject *)&Wrappertype);
+  return newWrapper(self,args,(PyTypeObject *)&Wrappertype);
 }
 
 static PyObject *
@@ -700,7 +873,7 @@ xaq_of(PyObject *self, PyObject *args)
       return NULL;
     }
 
-  return (PyObject*)newWrapper(self,args,(PyTypeObject *)&XaqWrappertype);
+  return newWrapper(self,args,(PyTypeObject *)&XaqWrappertype);
 }
 
 static struct PyMethodDef Acquirer_methods[] = {
@@ -721,7 +894,7 @@ void
 initAcquisition()
 {
   PyObject *m, *d;
-  char *rev="$Revision: 1.8 $";
+  char *rev="$Revision: 1.24 $";
   PURE_MIXIN_CLASS(Acquirer,
     "Base class for objects that implicitly"
     " acquire attributes from containers\n"
@@ -733,11 +906,15 @@ initAcquisition()
 
   UNLESS(ExtensionClassImported) return;
 
+  UNLESS(Acquired=PyString_FromStringAndSize(NULL,42)) return;
+  strcpy(PyString_AsString(Acquired),
+	 "<Special Object Used to Force Acquisition>");
+
   /* Create the module and add the functions */
   m = Py_InitModule4("Acquisition", methods,
-		     "Provide base classes for acquiring objects\n\n"
-		     "$Id: Acquisition.c,v 1.8 1997/07/02 20:15:27 jim Exp $\n",
-		     (PyObject*)NULL,PYTHON_API_VERSION);
+	   "Provide base classes for acquiring objects\n\n"
+	   "$Id: Acquisition.c,v 1.24 1998/11/17 19:49:43 jim Exp $\n",
+		     OBJECT(NULL),PYTHON_API_VERSION);
 
   d = PyModule_GetDict(m);
   init_py_names();
@@ -752,6 +929,7 @@ initAcquisition()
   
   PyDict_SetItemString(d,"__version__",
 		       PyString_FromStringAndSize(rev+11,strlen(rev+11)-2));
+  PyDict_SetItemString(d,"Acquired",Acquired);
 
   CHECK_FOR_ERRORS("can't initialize module Acquisition");
 }
